# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Admin views for invenio-accounts."""

from flask import current_app, flash
from flask_admin.actions import action
from flask_admin.babel import lazy_gettext
from flask_admin.contrib.sqla import ModelView
from flask_admin.contrib.sqla.ajax import QueryAjaxModelLoader
from flask_admin.form.fields import DateTimeField
from flask_admin.model.fields import AjaxSelectMultipleField
from flask_babelex import gettext as _
from flask_security.recoverable import send_reset_password_instructions
from flask_security.utils import hash_password
from invenio_db import db
from passlib import pwd
from werkzeug.local import LocalProxy
from wtforms.fields import BooleanField
from wtforms.validators import DataRequired

from .cli import commit
from .models import Role, SessionActivity, User, UserIdentity
from .sessions import delete_session

_datastore = LocalProxy(lambda: current_app.extensions['security'].datastore)


class UserView(ModelView):
    """Flask-Admin view to manage users."""

    can_view_details = True
    can_delete = False

    list_all = ('id', 'email', 'active', 'confirmed_at')

    column_list = \
        column_searchable_list = \
        column_sortable_list = \
        column_details_list = \
        list_all

    form_columns = ('email', 'password', 'active', 'roles', 'notification')

    form_args = dict(
        email=dict(label='Email', validators=[DataRequired()]),
        password=dict(default=lambda: pwd.genword(length=12))
    )

    form_extra_fields = {
        'notification': BooleanField(
            'Send User Notification',
            description='Send the new user an email about their account.')
    }

    column_filters = ('id', 'email', 'active', 'confirmed_at')

    column_default_sort = ('email', True)

    def on_model_change(self, form, User, is_created):
        """Hash password when saving."""
        if form.password.data is not None:
            pwd_ctx = current_app.extensions['security'].pwd_context
            if pwd_ctx.identify(form.password.data) is None:
                User.password = hash_password(form.password.data)

    def after_model_change(self, form, User, is_created):
        """Send password instructions if desired."""
        if is_created and form.notification.data is True:
            send_reset_password_instructions(User)

    @action('inactivate', _('Inactivate'),
            _('Are you sure you want to inactivate selected users?'))
    @commit
    def action_inactivate(self, ids):
        """Inactivate users."""
        try:
            count = 0
            for user_id in ids:
                user = _datastore.get_user(user_id)
                if user is None:
                    raise ValueError(_("Cannot find user."))
                if _datastore.deactivate_user(user):
                    count += 1
            if count > 0:
                flash(_('User(s) were successfully inactivated.'), 'success')
        except Exception as exc:
            if not self.handle_view_exception(exc):
                raise

            current_app.logger.exception(str(exc))  # pragma: no cover
            flash(_('Failed to inactivate users.'),
                  'error')  # pragma: no cover

    @action('activate', _('Activate'),
            _('Are you sure you want to activate selected users?'))
    @commit
    def action_activate(self, ids):
        """Inactivate users."""
        try:
            count = 0
            for user_id in ids:
                user = _datastore.get_user(user_id)
                if user is None:
                    raise ValueError(_("Cannot find user."))
                if _datastore.activate_user(user):
                    count += 1
            if count > 0:
                flash(_('User(s) were successfully inactivated.'), 'success')
        except Exception as exc:
            if not self.handle_view_exception(exc):
                raise

            current_app.logger.exception(str(exc))  # pragma: no cover
            flash(_('Failed to activate users.'), 'error')  # pragma: no cover


class RoleView(ModelView):
    """Admin view for roles."""

    can_view_details = True

    list_all = ('id', 'name', 'description')

    column_list = \
        column_searchable_list = \
        column_filters = \
        column_details_list = \
        columns_sortable_list = \
        list_all

    form_columns = ('name', 'description', 'users')

    user_loader = QueryAjaxModelLoader(
        'user',
        LocalProxy(lambda: _datastore.db.session),
        User,
        fields=['email'],
        page_size=10
    )

    form_extra_fields = {
        'users': AjaxSelectMultipleField(user_loader)
    }

    form_ajax_refs = {
        'user': user_loader
    }


class SessionActivityView(ModelView):
    """Admin view for user sessions."""

    can_view_details = False
    can_create = False
    can_edit = False

    list_all = ('user.id', 'user.email', 'sid_s', 'created')

    column_labels = {
        'user.id': 'User ID',
        'user.email': 'Email',
        'sid_s': 'Session ID',
    }
    column_list = list_all
    column_filters = list_all
    column_sortable_list = list_all

    form_ajax_refs = {
        'user': {
            'fields': (User.id, User.email)
        },
    }

    def delete_model(self, model):
        """Delete a specific session."""
        if SessionActivity.is_current(sid_s=model.sid_s):
            flash('You could not remove your current session', 'error')
            return
        delete_session(sid_s=model.sid_s)
        db.session.commit()

    @action('delete', lazy_gettext('Delete selected sessions'),
            lazy_gettext('Are you sure you want to delete selected sessions?'))
    def action_delete(self, ids):
        """Delete selected sessions."""
        is_current = any(SessionActivity.is_current(sid_s=id_) for id_ in ids)
        if is_current:
            flash('You could not remove your current session', 'error')
            return
        for id_ in ids:
            delete_session(sid_s=id_)
        db.session.commit()


class UserIdentityView(ModelView):
    """Flask-Admin view to manage user identities from invenio-accounts."""

    can_view_details = True
    can_create = False

    column_list = (
        'id',
        'method',
        'id_user',
        'user.email',
    )

    column_searchable_list = \
        column_sortable_list = \
        column_list

    column_filters = (
        'id',
        'method',
        'id_user',
        'user.email',
    )

    column_labels = {
        'user.email': _("Email"),
        'id_user': _("User ID"),
    }


session_adminview = {
    'model': SessionActivity,
    'modelview': SessionActivityView,
    'category': _('User Management')
}

user_adminview = {
    'model': User,
    'modelview': UserView,
    'category': _('User Management')
}

role_adminview = {
    'model': Role,
    'modelview': RoleView,
    'category': _('User Management')
}

user_identity_adminview = {
    'model': UserIdentity,
    'modelview': UserIdentityView,
    'category': _('User Management'),
    'name': _('Linked account identities'),
}

__all__ = (
    'role_adminview',
    'RoleView',
    'session_adminview',
    'SessionActivityView',
    'user_adminview',
    'user_identity_adminview',
    'UserView',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create accounts tables."""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from sqlalchemy.engine.reflection import Inspector

# revision identifiers, used by Alembic.
revision = '9848d0149abd'
down_revision = '843bc79c426f'
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.create_table(
        'accounts_role',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(length=80), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name')
    )
    op.create_table(
        'accounts_user',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('email', sa.String(length=255), nullable=True),
        sa.Column('password', sa.String(length=255), nullable=True),
        sa.Column('active', sa.Boolean(name='active'), nullable=True),
        sa.Column('confirmed_at', sa.DateTime(), nullable=True),
        sa.Column('last_login_at', sa.DateTime(), nullable=True),
        sa.Column('current_login_at', sa.DateTime(), nullable=True),
        sa.Column('last_login_ip',
                  sqlalchemy_utils.types.ip_address.IPAddressType(),
                  nullable=True),
        sa.Column('current_login_ip',
                  sqlalchemy_utils.types.ip_address.IPAddressType(),
                  nullable=True),
        sa.Column('login_count', sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('email')
    )
    op.create_table(
        'accounts_user_session_activity',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column('sid_s', sa.String(length=255), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ['user_id'], [u'accounts_user.id'],
            name='fk_accounts_session_activity_user_id',
        ),
        sa.PrimaryKeyConstraint('sid_s')
    )
    op.create_table(
        'accounts_userrole',
        sa.Column('user_id', sa.Integer(), nullable=True),
        sa.Column('role_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ['role_id'], ['accounts_role.id'],
            name='fk_accounts_userrole_role_id',
        ),
        sa.ForeignKeyConstraint(
            ['user_id'], ['accounts_user.id'],
            name='fk_accounts_userrole_user_id',
        ),
    )
    with op.batch_alter_table('transaction') as batch_op:
        batch_op.add_column(sa.Column(
            'user_id',
            sa.Integer(),
            sa.ForeignKey('accounts_user.id'),
            nullable=True,
        ))
        batch_op.create_index(
            op.f('ix_transaction_user_id'), ['user_id'], unique=False
        )


def downgrade():
    """Downgrade database."""
    ctx = op.get_context()
    insp = Inspector.from_engine(ctx.connection.engine)

    for fk in insp.get_foreign_keys('transaction'):
        if fk['referred_table'] == 'accounts_user':
            op.drop_constraint(
                op.f(fk['name']), 'transaction', type_='foreignkey'
            )

    with op.batch_alter_table('transaction') as batch_op:
        batch_op.drop_index(op.f('ix_transaction_user_id'))
        batch_op.drop_column('user_id')
    op.drop_table('accounts_userrole')
    op.drop_table('accounts_user_session_activity')
    op.drop_table('accounts_user')
    op.drop_table('accounts_role')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Session-aware datastore."""

from flask_security import SQLAlchemyUserDatastore

from .sessions import delete_user_sessions


class SessionAwareSQLAlchemyUserDatastore(SQLAlchemyUserDatastore):
    """Datastore which deletes active session when a user is deactivated."""

    def deactivate_user(self, user):
        """Deactivate a  user.

        :param user: A :class:`invenio_accounts.models.User` instance.
        :returns: The datastore instance.
        """
        res = super(SessionAwareSQLAlchemyUserDatastore, self).deactivate_user(
            user)
        if res:
            delete_user_sessions(user)
        return res



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Pytest configuration."""

import os
import shutil
import tempfile

import pytest
from flask import Flask
from flask.cli import ScriptInfo
from flask_admin import Admin
from flask_babelex import Babel
from flask_celeryext import FlaskCeleryExt
from flask_mail import Mail
from flask_menu import Menu
from invenio_db import InvenioDB, db
from invenio_i18n import InvenioI18N
from invenio_rest import InvenioREST
from simplekv.memory.redisstore import RedisStore
from sqlalchemy_utils.functions import create_database, database_exists, \
    drop_database

from invenio_accounts import InvenioAccounts, InvenioAccountsREST
from invenio_accounts.admin import role_adminview, session_adminview, \
    user_adminview
from invenio_accounts.testutils import create_test_user
from invenio_accounts.views.rest import create_blueprint


def _app_factory(config=None):
    """Application factory."""
    # TODO use the fixtures from pytest-invenio instead
    instance_path = tempfile.mkdtemp()
    app = Flask('testapp', instance_path=instance_path)
    icons = {
        'semantic-ui': {
            'key': 'key icon',
            'link': 'linkify icon',
            'shield': 'shield alternate icon',
            'user': 'user icon',
            'codepen': 'codepen icon',
            'cogs': 'cogs icon',
            '*': '{} icon'
        },
        'bootstrap3': {
            'key': 'fa fa-key fa-fw',
            'link': 'fa fa-link fa-fw',
            'shield': 'fa fa-shield fa-fw',
            'user': 'fa fa-user fa-fw',
            'codepen': 'fa fa-codepen fa-fw',
            'cogs': 'fa fa-cogs fa-fw',
            '*': 'fa fa-{} fa-fw',
        }
    }

    app.config.update(
        ACCOUNTS_USE_CELERY=False,
        ACCOUNTS_LOCAL_LOGIN_ENABLED=True,
        APP_THEME=["semantic-ui"],
        THEME_ICONS=icons,
        CELERY_ALWAYS_EAGER=True,
        CELERY_CACHE_BACKEND="memory",
        CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
        CELERY_RESULT_BACKEND="cache",
        LOGIN_DISABLED=False,
        MAIL_SUPPRESS_SEND=True,
        SECRET_KEY="CHANGE_ME",
        SECURITY_PASSWORD_SALT="CHANGE_ME_ALSO",
        SECURITY_CONFIRM_EMAIL_WITHIN="2 seconds",
        SECURITY_RESET_PASSWORD_WITHIN="2 seconds",
        SECURITY_CHANGEABLE=True,
        SECURITY_RECOVERABLE=True,
        SECURITY_REGISTERABLE=True,
        SQLALCHEMY_DATABASE_URI=os.environ.get(
            'SQLALCHEMY_DATABASE_URI', 'sqlite:///test.db'),
        SERVER_NAME='example.com',
        TESTING=True,
        WTF_CSRF_ENABLED=False,
    )

    app.config.update(config or {})
    Menu(app)
    Babel(app)
    Mail(app)
    InvenioDB(app)
    return app


def _database_setup(app, request):
    """Set up the database."""
    with app.app_context():
        if not database_exists(str(db.engine.url)):
            create_database(str(db.engine.url))
        db.create_all()

    def teardown():
        with app.app_context():
            if database_exists(str(db.engine.url)):
                drop_database(str(db.engine.url))
            # Delete sessions in kvsession store
            if hasattr(app, 'kvsession_store') and \
                    isinstance(app.kvsession_store, RedisStore):
                app.kvsession_store.redis.flushall()
        shutil.rmtree(app.instance_path)

    request.addfinalizer(teardown)
    return app


@pytest.fixture()
def base_app(request):
    """Flask application fixture."""
    app = _app_factory()
    _database_setup(app, request)
    yield app


@pytest.fixture()
def app(request):
    """Flask application fixture with Invenio Accounts."""
    app = _app_factory()
    app.config.update(ACCOUNTS_USERINFO_HEADERS=True)
    InvenioAccounts(app)

    from invenio_accounts.views.settings import blueprint
    app.register_blueprint(blueprint)

    _database_setup(app, request)
    yield app


@pytest.fixture()
def api(request):
    """Flask application fixture."""
    api_app = _app_factory(
        dict(
            SQLALCHEMY_DATABASE_URI=os.environ.get(
                'SQLALCHEMY_DATABASE_URI', 'sqlite:///test.db'),
            SERVER_NAME='localhost',
            TESTING=True,
        ))

    InvenioREST(api_app)
    InvenioAccountsREST(api_app)
    api_app.register_blueprint(create_blueprint(api_app))

    _database_setup(api_app, request)

    yield api_app


@pytest.fixture()
def app_with_redis_url(request):
    """Flask application fixture with Invenio Accounts."""
    app = _app_factory(dict(
        ACCOUNTS_SESSION_REDIS_URL='redis://localhost:6379/0'
    ))
    app.config.update(ACCOUNTS_USERINFO_HEADERS=True)
    InvenioAccounts(app)

    from invenio_accounts.views.settings import blueprint
    app.register_blueprint(blueprint)

    _database_setup(app, request)
    yield app


@pytest.fixture()
def app_with_flexible_registration(request):
    """Flask application fixture with Invenio Accounts."""
    from webargs import fields

    from invenio_accounts.views.rest import RegisterView, use_kwargs

    class MyRegisterView(RegisterView):

        post_args = {
            **RegisterView.post_args,
            'active': fields.Boolean(required=True)
        }

        @use_kwargs(post_args)
        def post(self, **kwargs):
            """Register a user."""
            return super(MyRegisterView, self).post(**kwargs)

    api_app = _app_factory()
    InvenioREST(api_app)
    InvenioAccountsREST(api_app)

    api_app.config['ACCOUNTS_REST_AUTH_VIEWS']['register'] = MyRegisterView

    api_app.register_blueprint(create_blueprint(api_app))

    _database_setup(api_app, request)
    yield api_app


@pytest.fixture
def script_info(app):
    """Get ScriptInfo object for testing CLI."""
    return ScriptInfo(create_app=lambda info: app)


@pytest.fixture
def task_app(request):
    """Flask application with Celery enabled."""
    app = _app_factory(dict(
        ACCOUNTS_USE_CELERY=True,
        MAIL_SUPPRESS_SEND=True,
    ))
    FlaskCeleryExt(app)
    InvenioAccounts(app)
    _database_setup(app, request)
    return app


@pytest.fixture
def cookie_app(request):
    """Flask application  enabled."""
    app = _app_factory(dict(
        SESSION_COOKIE_SECURE=True,
        SESSION_COOKIE_DOMAIN='example.com',
    ))
    InvenioAccounts(app)
    _database_setup(app, request)
    return app


@pytest.fixture
def admin_view(app):
    """Admin view fixture."""
    assert isinstance(role_adminview, dict)
    assert isinstance(user_adminview, dict)
    assert isinstance(session_adminview, dict)

    assert 'model' in role_adminview
    assert 'modelview' in role_adminview
    assert 'model' in user_adminview
    assert 'modelview' in user_adminview
    assert 'model' in session_adminview
    assert 'modelview' in session_adminview

    admin = Admin(app, name="Test")

    user_adminview_copy = dict(user_adminview)
    user_model = user_adminview_copy.pop('model')
    user_view = user_adminview_copy.pop('modelview')
    admin.add_view(user_view(user_model, db.session, **user_adminview_copy))

    admin.add_view(session_adminview['modelview'](
        session_adminview['model'], db.session,
        category=session_adminview['category']))


@pytest.fixture()
def app_i18n(app):
    """Init invenio-i18n."""
    InvenioI18N(app)
    return app


@pytest.fixture()
def users(app):
    """Create users."""
    user1 = create_test_user(email='info@inveniosoftware.org',
                             password='tester')
    user2 = create_test_user(email='info2@inveniosoftware.org',
                             password='tester2')

    return [
        {'email': user1.email, 'id': user1.id,
         'password': user1.password_plaintext, 'obj': user1},
        {'email': user2.email, 'id': user2.id,
         'password': user2.password_plaintext, 'obj': user2},
    ]



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE

"""Test authentication REST API views."""

import datetime
import json
from unittest import mock

from flask import url_for
from flask_security import current_user
from invenio_db import db

from invenio_accounts.testutils import create_test_user


def get_json(response):
    """Get JSON from response."""
    return json.loads(response.get_data(as_text=True))


def assert_error_resp(res, expected_errors, expected_status_code=400):
    """Assert errors in a client response."""
    assert res.status_code == expected_status_code
    payload = get_json(res)
    errors = payload.get('errors', [])
    for field, msg in expected_errors:
        if not field:  # top-level "message" error
            assert msg in payload['message'].lower()
            continue
        assert any(
            e['field'] == field and msg in e['message'].lower()
            for e in errors), payload


def _mock_send_mail(subject, recipient, template, **context):
    from urllib.parse import urlsplit

    from invenio_accounts.config import SECURITY_RESET_URL

    assert context['reset_link']
    assert SECURITY_RESET_URL in urlsplit(context['reset_link'])[2]


def _mock_send_confirmation_mail(subject, recipient, template, **context):
    from invenio_accounts.config import ACCOUNTS_REST_CONFIRM_EMAIL_ENDPOINT

    assert context['confirmation_link']
    assert ACCOUNTS_REST_CONFIRM_EMAIL_ENDPOINT.format(token='') in \
        context['confirmation_link']


def _login_user(client, user, email='normal@test.com', password='123456'):
    url = url_for('invenio_accounts_rest_auth.login')
    res = client.post(url, data=dict(email=email, password=password))
    payload = get_json(res)
    assert res.status_code == 200
    assert payload['id'] == user.id
    assert payload['email'] == user.email
    session_cookie = next(
        c for c in client.cookie_jar if c.name == 'session')
    assert session_cookie is not None
    assert session_cookie.value
    assert current_user.is_authenticated


def test_login_view(api):
    app = api
    with app.app_context():

        normal_user = create_test_user(email='normal@test.com')
        create_test_user(email='disabled@test.com', active=False)
        create_test_user(email='nopass@test.com', password=None)

        db.session.commit()

        with app.test_client() as client:
            url = url_for('invenio_accounts_rest_auth.login')
            # Missing fields
            res = client.post(url)
            assert_error_resp(res, (
                ('password', 'required'),
                ('email', 'required'),
            ))

            # Invalid fields
            res = client.post(url, data=dict(
                email='invalid-email', password=None))
            assert_error_resp(res, (
                ('password', 'required field'),
                ('email', 'not a valid email'),
            ))

            # Invalid credentials
            res = client.post(url, data=dict(
                email='not@exists.com', password='123456'))
            assert_error_resp(res, (
                ('email', 'user does not exist'),
            ))

            # No-password user
            res = client.post(url, data=dict(
                email='nopass@test.com', password='123456'))
            assert_error_resp(res, (
                ('password', 'no password is set'),
            ))

            # Disabled account
            res = client.post(url, data=dict(
                email='disabled@test.com', password='123456'))
            assert_error_resp(res, (
                (None, 'account is disabled'),
            ))

            # Successful login
            _login_user(client, normal_user)

            # User info view
            res = client.get(url_for('invenio_accounts_rest_auth.user_info'))
            payload = get_json(res)
            assert res.status_code == 200
            assert payload['id'] == normal_user.id


def test_disabled_login_view(api):
    app = api
    app.config["ACCOUNTS_LOCAL_LOGIN_ENABLED"] = False

    with app.app_context():
        user = create_test_user(email='normal@test.com')
        db.session.commit()
        with app.test_client() as client:
            url = url_for('invenio_accounts_rest_auth.login')
            data = {
                "email": user.email,
                "password": "123456"
            }

            res = client.post(url, data=data)
            payload = get_json(res)

            assert res.status_code != 200
            expected_msg = app.config["SECURITY_MSG_LOCAL_LOGIN_DISABLED"][0]
            assert payload["message"] == expected_msg


def test_registration_view(api):
    app = api
    with app.app_context():
        create_test_user(email='old@test.com')
        db.session.commit()
        with app.test_client() as client:
            url = url_for('invenio_accounts_rest_auth.register')

            # Missing fields
            res = client.post(url)
            assert_error_resp(res, (
                ('password', 'required'),
                ('email', 'required'),
            ))

            # Invalid fields
            res = client.post(url, data=dict(
                email='invalid-email', password='short'))
            assert_error_resp(res, (
                ('password', 'length'),
                ('email', 'not a valid email'),
            ))

            # Existing user
            res = client.post(url, data=dict(
                email='old@test.com', password='123456'))
            assert_error_resp(res, (
                ('email', 'old@test.com is already associated'),
            ))

            # Successful registration
            res = client.post(url, data=dict(
                email='new@test.com', password='123456'))
            payload = get_json(res)
            assert res.status_code == 200
            assert payload['id'] == 2
            assert payload['email'] == 'new@test.com'
            session_cookie = next(
                c for c in client.cookie_jar if c.name == 'session')
            assert session_cookie is not None
            assert session_cookie.value

            res = client.get(url_for('invenio_accounts_rest_auth.user_info'))
            assert res.status_code == 200


def test_custom_registration_view(app_with_flexible_registration):
    app = app_with_flexible_registration
    with app.app_context():
        create_test_user(email='old@test.com')
        db.session.commit()
        with app.test_client() as client:
            url = url_for('invenio_accounts_rest_auth.register')

            # Missing custom field
            res = client.post(url, data=dict(
                email='new@test.com', password='123456'))
            assert_error_resp(res, (
                ('active', 'required'),
            ))

            # Successful registration
            res = client.post(url, data=dict(
                email='new@test.com', password='123456', active=True))
            assert res.status_code == 200


def test_logout_view(api):
    app = api
    with app.app_context():
        normal_user = create_test_user(email='normal@test.com')
        db.session.commit()
        with app.test_client() as client:

            # Login user
            _login_user(client, normal_user)
            old_session_cookie = next(
                c for c in client.cookie_jar if c.name == 'session')

            # Log out user
            url = url_for('invenio_accounts_rest_auth.logout')
            res = client.post(url)
            payload = get_json(res)
            assert payload['message'] == 'User logged out.'
            new_session_cookie = next(
                c for c in client.cookie_jar if c.name == 'session')
            assert old_session_cookie.value != new_session_cookie.value
            assert current_user.is_anonymous


@mock.patch('invenio_accounts.views.rest.send_mail', _mock_send_mail)
def test_forgot_password_view(api):
    app = api
    with app.app_context():
        normal_user = create_test_user(email='normal@test.com')
        db.session.commit()
        with app.test_client() as client:
            url = url_for('invenio_accounts_rest_auth.forgot_password')

            # Invalid email
            res = client.post(url, data=dict(email='invalid'))
            assert_error_resp(res, (
                ('email', 'user does not exist'),
            ))

            # Valid email
            res = client.post(url, data=dict(email='normal@test.com'))
            payload = get_json(res)
            assert 'Instructions to reset your password' in payload['message']


def test_disabled_forgot_password_view(api):
    from flask_security.recoverable import generate_reset_password_token
    app = api
    app.config["SECURITY_RECOVERABLE"] = False
    app.extensions["security"].recoverable = False

    with app.app_context():
        user = create_test_user(email='normal@test.com')
        token = generate_reset_password_token(user)
        db.session.commit()
        with app.test_client() as client:
            url = url_for('invenio_accounts_rest_auth.forgot_password')
            res = client.post(url, data={"email": user.email})
            payload = get_json(res)

            assert res.status_code != 200
            expected = app.config["SECURITY_MSG_PASSWORD_RECOVERY_DISABLED"][0]
            assert payload["message"] == expected


def test_reset_password_view(api):
    from flask_security.recoverable import generate_reset_password_token
    app = api
    with app.app_context():
        normal_user = create_test_user(email='normal@test.com')
        # Generate token
        token = generate_reset_password_token(normal_user)
        db.session.commit()
        with app.test_client() as client:
            url = url_for('invenio_accounts_rest_auth.reset_password')
            res = client.post(url, data=dict(password='new-password',
                                             token=token))
            payload = get_json(res)
            assert res.status_code == 200
            assert 'You successfully reset your password' in payload['message']

            # Login using new password
            res = client.post(url, data=dict(
                email='normal@test.com', password='new-password'))
            payload = get_json(res)
            assert res.status_code == 200
            assert payload['id'] == normal_user.id
            assert payload['email'] == normal_user.email
            session_cookie = next(
                c for c in client.cookie_jar if c.name == 'session')
            assert session_cookie is not None
            assert session_cookie.value


def test_disabled_reset_password_view(api):
    from flask_security.recoverable import generate_reset_password_token
    app = api
    app.config["SECURITY_RECOVERABLE"] = False
    app.extensions["security"].recoverable = False

    with app.app_context():
        user = create_test_user(email='normal@test.com')
        token = generate_reset_password_token(user)
        db.session.commit()
        with app.test_client() as client:
            url = url_for('invenio_accounts_rest_auth.reset_password')
            data = {
                "token": token,
                "password": "123456",
            }

            res = client.post(url, data=data)
            payload = get_json(res)

            assert res.status_code != 200
            expected = app.config["SECURITY_MSG_PASSWORD_RESET_DISABLED"][0]
            assert payload["message"] == expected


def test_change_password_view(api):
    app = api
    with app.app_context():
        normal_user = create_test_user(email='normal@test.com')
        db.session.commit()
        with app.test_client() as client:
            url = url_for('invenio_accounts_rest_auth.change_password')

            # Not authorized user
            res = client.post(url, data=dict(password='123456',
                                             new_password='new-password'))
            assert_error_resp(res, (
                (None, 'server could not verify that you are authorized'),
            ), expected_status_code=401)

            # Logged in user
            _login_user(client, normal_user)

            # Same password
            res = client.post(url, data=dict(password='123456',
                                             new_password='123456'))
            assert_error_resp(res, (
                ('password', 'new password must be different'),
            ))

            # Valid password
            res = client.post(url, data=dict(password='123456',
                                             new_password='new-password'))
            payload = get_json(res)
            assert 'successfully changed your password' in payload['message']
            # Log in using new password
            _login_user(client, normal_user, password='new-password')


def test_disabled_change_password_view(api):
    app = api
    app.config["SECURITY_CHANGEABLE"] = False
    app.extensions["security"].changeable = False

    with app.app_context():
        user = create_test_user(email='normal@test.com')
        db.session.commit()
        with app.test_client() as client:
            _login_user(client, user)
            url = url_for('invenio_accounts_rest_auth.change_password')
            data = {
                "password": user.password_plaintext,
                "new_password": "1234567",
            }

            res = client.post(url, data=data)
            payload = get_json(res)

            assert res.status_code != 200
            expected = app.config["SECURITY_MSG_PASSWORD_CHANGE_DISABLED"][0]
            assert payload["message"] == expected


@mock.patch('invenio_accounts.views.rest.send_mail',
            _mock_send_confirmation_mail)
def test_send_confirmation_email_view(api):
    app = api
    with app.app_context():
        normal_user = create_test_user(email='normal@test.com')
        confirmed_user = create_test_user(email='confirmed@test.com',
                                          confirmed_at=datetime.datetime.now())

        db.session.commit()
        with app.test_client() as client:
            url = url_for('invenio_accounts_rest_auth.send_confirmation')
            _login_user(client, normal_user)
            res = client.post(url, data=dict(email=normal_user.email))

            assert_error_resp(res, (
                (None, 'confirmation instructions have been sent'),
            ), expected_status_code=200)

            # Already confirmed
            res = client.post(url, data=dict(email=confirmed_user.email))
            assert_error_resp(res, (
                (None, 'email has already been confirmed'),
            ))


def test_confirm_email_view(api):
    from flask_security.confirmable import generate_confirmation_token
    app = api
    with app.app_context():
        normal_user = create_test_user(email='normal@test.com')
        confirmed_user = create_test_user(email='confirmed@test.com',
                                          confirmed_at=datetime.datetime.now())

        db.session.commit()
        # Generate token
        token = generate_confirmation_token(normal_user)
        confirmed_token = generate_confirmation_token(confirmed_user)
        with app.test_client() as client:
            url = url_for('invenio_accounts_rest_auth.confirm_email')

            # Invalid token
            res = client.post(url, data=dict(token='foo'))
            payload = get_json(res)
            assert 'invalid confirmation token' in \
                payload['message'][0].lower()

            # Already confirmed user
            res = client.post(url, data=dict(token=confirmed_token))
            payload = get_json(res)
            assert 'email has already been confirmed' in \
                payload['message'].lower()

            # Valid confirm user
            assert normal_user.confirmed_at is None
            res = client.post(url, data=dict(token=token))
            payload = get_json(res)
            assert 'your email has been confirmed' in \
                payload['message'].lower()
            assert normal_user.confirmed_at


def test_sessions_list_view(api):
    app = api
    session_total = 3
    with app.app_context():
        normal_user = create_test_user(email='normal@test.com')
        db.session.commit()

        url = url_for('invenio_accounts_rest_auth.sessions_list')

        for _ in range(session_total):
            with app.test_client() as client:
                _login_user(client, normal_user)

        res = client.get(url)
        payload = get_json(res)
        assert payload['total'] == session_total
        assert isinstance(payload['results'], list)


def test_sessions_item_view(api):
    app = api
    with app.app_context():
        normal_user = create_test_user(email='normal@test.com')
        db.session.commit()

        old_sid_s = None
        with app.test_client() as client:
            # Invalid session
            _login_user(client, normal_user)
            url = url_for('invenio_accounts_rest_auth.sessions_item',
                          sid_s='foo')
            res = client.delete(url)
            assert_error_resp(res, (
                (None, 'unable to remove session'),
            ))

            # Valid session
            sid_s = current_user.active_sessions[0].sid_s
            assert current_user
            url = url_for('invenio_accounts_rest_auth.sessions_item',
                          sid_s=sid_s)
            res = client.delete(url)
            assert_error_resp(res, (
                (None, 'successfully removed'),
                (None, 'logged out'),
            ), expected_status_code=200)
            assert not current_user.active_sessions

            # Login again and save session
            _login_user(client, normal_user)
            old_sid_s = current_user.active_sessions[0].sid_s

        with app.test_client() as client:
            _login_user(client, normal_user)
            url = url_for('invenio_accounts_rest_auth.sessions_item',
                          sid_s=old_sid_s)
            res = client.delete(url)
            assert_error_resp(res, (
                (None, 'successfully removed'),
                (None, 'revoked'),
            ), expected_status_code=200)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Legacy Invenio hash support."""

import hashlib

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from passlib.utils.compat import str_to_uascii
from passlib.utils.handlers import GenericHandler, HasSalt, parse_mc2, \
    render_mc2

__all__ = ('InvenioAesEncryptedEmail', )


def _to_binary(val):
    """Convert to binary."""
    if isinstance(val, str):
        return val.encode('utf-8')
    assert isinstance(val, bytes)
    return val


def _to_string(val):
    """Convert to text."""
    if isinstance(val, bytes):
        return val.decode('utf-8')
    assert isinstance(val, str)
    return val


def _mysql_aes_key(key):
    """Format key."""
    final_key = bytearray(16)
    for i, c in enumerate(key):
        final_key[i % 16] ^= key[i]
    return bytes(final_key)


def _mysql_aes_pad(val):
    """Padding."""
    val = _to_string(val)
    pad_value = 16 - (len(val) % 16)
    return _to_binary('{0}{1}'.format(val, chr(pad_value) * pad_value))


def _mysql_aes_unpad(val):
    """Reverse padding."""
    val = _to_string(val)
    pad_value = ord(val[-1])
    return val[:-pad_value]


def _mysql_aes_engine(key):
    """Create MYSQL AES cipher engine."""
    return Cipher(algorithms.AES(key), modes.ECB(), default_backend())


def mysql_aes_encrypt(val, key):
    """Mysql AES encrypt value with secret key.

    :param val: Plain text value.
    :param key: The AES key.
    :returns: The encrypted AES value.
    """
    assert isinstance(val, bytes) or isinstance(val, str)
    assert isinstance(key, bytes) or isinstance(key, str)
    k = _mysql_aes_key(_to_binary(key))
    v = _mysql_aes_pad(_to_binary(val))
    e = _mysql_aes_engine(k).encryptor()

    return e.update(v) + e.finalize()


def mysql_aes_decrypt(encrypted_val, key):
    """Mysql AES decrypt value with secret key.

    :param encrypted_val: Encrypted value.
    :param key: The AES key.
    :returns: The AES value decrypted.
    """
    assert isinstance(encrypted_val, bytes) \
        or isinstance(encrypted_val, str)
    assert isinstance(key, bytes) or isinstance(key, str)
    k = _mysql_aes_key(_to_binary(key))
    d = _mysql_aes_engine(_to_binary(k)).decryptor()
    return _mysql_aes_unpad(d.update(_to_binary(encrypted_val)) + d.finalize())


class InvenioAesEncryptedEmail(HasSalt, GenericHandler):
    """Invenio AES encryption of user email using password as secret key.

    Invenio 1.x was AES encrypting the users email address with the password
    as the secret key and storing it in a blob column. This e.g. caused
    problems when a user wanted to change email address.
    This hashing engine, differs from Invenio 1.x in that it sha256 hashes the
    encrypted value as well to produce a string in the same length instead of
    a binary blob. It is not done for extra security, just for convenience of
    migration to using passlib's sha512.
    An upgrade recipe is provided to migrated existing binary password hashes
    to hashes of this engine.
    """

    name = "invenio_aes_encrypted_email"
    setting_kwds = "salt"
    ident = u"$invenio-aes$"

    @classmethod
    def from_string(cls, hash, **context):
        """Parse instance from configuration string in Modular Crypt Format."""
        salt, checksum = parse_mc2(hash, cls.ident, handler=cls)

        return cls(salt=salt, checksum=checksum)

    def to_string(self):
        """Render instance to configuration string in Modular Crypt Format."""
        return render_mc2(self.ident, self.salt, self.checksum)

    def _calc_checksum(self, secret):
        """Calculate string.

        :param secret: The secret key.
        :returns: The checksum.
        """
        return str_to_uascii(
            hashlib.sha256(mysql_aes_encrypt(self.salt, secret)).hexdigest()
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C)      2021 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio user management and authentication."""

from warnings import warn

import pkg_resources
from flask import Blueprint, abort, current_app, request_finished, session
from flask_kvsession import KVSessionExtension
from flask_login import LoginManager, user_logged_in, user_logged_out
from flask_principal import AnonymousIdentity
from flask_security import Security
from invenio_db import db
from passlib.registry import register_crypt_handler
from werkzeug.utils import cached_property

from invenio_accounts.forms import confirm_register_form_factory, \
    login_form_factory, register_form_factory

from . import config
from .datastore import SessionAwareSQLAlchemyUserDatastore
from .hash import InvenioAesEncryptedEmail
from .models import Role, User
from .sessions import csrf_token_reset, login_listener, logout_listener
from .utils import obj_or_import_string, set_session_info


class InvenioAccounts(object):
    """Invenio-Accounts extension."""

    def __init__(self, app=None, sessionstore=None):
        """Extension initialization.

        :param app: The Flask application.
        :param sessionstore: store for sessions. Passed to
            ``flask-kvsession``. Defaults to redis.
        """
        self.security = Security()
        self.datastore = None
        if app:
            self.init_app(app, sessionstore=sessionstore)

    @staticmethod
    def monkey_patch_flask_security():
        """Monkey-patch Flask-Security."""
        # Disable remember me cookie generation as it does not work with
        # session activity tracking (a remember me token will bypass revoking
        # of a session).
        def patch_do_nothing(*args, **kwargs):
            pass
        LoginManager._set_cookie = patch_do_nothing

    @cached_property
    def jwt_decode_factory(self):
        """Load default JWT veryfication factory."""
        return obj_or_import_string(
            current_app.config.get('ACCOUNTS_JWT_DECODE_FACTORY'))

    @cached_property
    def jwt_creation_factory(self):
        """Load default JWT creation factory."""
        return obj_or_import_string(
            current_app.config.get('ACCOUNTS_JWT_CREATION_FACTORY'))

    def register_anonymous_identity_loader(self, state):
        """Registers a loader for AnonymousIdentity.

        Additional loader is necessary for applying a need 'any-user' to
        AnonymousUser in the invenio-access module
        """
        # Attention: the order of the loaders is important
        # append is used here instead of decorator to enforce the order
        state.principal.identity_loaders.append(AnonymousIdentity)

    def check_configuration_consistency(self, app):
        """Check if the config is consistent and issue a warning if not."""
        # Warn if inconsistent configuration is detected
        sec = app.extensions["security"]
        local_login = app.config.get("ACCOUNTS_LOCAL_LOGIN_ENABLED", True)
        local_account_editable = (
            sec.registerable or sec.changeable or sec.recoverable
        )

        if local_account_editable and not local_login:
            warning_message = (
                "ACCOUNTS_LOCAL_LOGIN_ENABLED is False, while at least one "
                "of SECURITY_REGISTERABLE, SECURITY_RECOVERABLE and "
                "SECURITY_CHANGEABLE is True"
            )
            warn(warning_message, Warning, stacklevel=2)

    def init_app(self, app, sessionstore=None, register_blueprint=True):
        """Flask application initialization.

        The following actions are executed:

        #. Initialize the configuration.

        #. Monkey-patch Flask-Security.

        #. Create the user datastore.

        #. Create the sessionstore.

        #. Initialize the extension, the forms to register users and
           confirms their emails, the CLI and, if ``ACCOUNTS_USE_CELERY`` is
           ``True``, register a celery task to send emails.

        #. Override Flask-Security's default login view function.

        #. Warn if inconsistent configuration is detected

        :param app: The Flask application.
        :param sessionstore: store for sessions. Passed to
            ``flask-kvsession``. If ``None`` then Redis is configured.
            (Default: ``None``)
        :param register_blueprint: If ``True``, the application registers the
            blueprints. (Default: ``True``)
        """
        self.init_config(app)

        # Monkey-patch Flask-Security
        InvenioAccounts.monkey_patch_flask_security()

        # Create user datastore
        if not self.datastore:
            self.datastore = SessionAwareSQLAlchemyUserDatastore(
                db, User, Role)

        if app.config['ACCOUNTS_SESSION_ACTIVITY_ENABLED']:
            self._enable_session_activity(app=app)

        # Initialize extension.
        _register_blueprint = app.config.get('ACCOUNTS_REGISTER_BLUEPRINT')
        if _register_blueprint is not None:
            register_blueprint = _register_blueprint

        state = self.security.init_app(app, datastore=self.datastore,
                                       register_blueprint=register_blueprint)

        # Override Flask-Security's default login view function
        new_login_view = obj_or_import_string(
            app.config.get("ACCOUNTS_LOGIN_VIEW_FUNCTION")
        )
        if new_login_view is not None:
            app.view_functions["security.login"] = new_login_view

        self.register_anonymous_identity_loader(state)

        app.extensions['security'].register_form = register_form_factory(
            app.extensions['security'].register_form, app)

        app.extensions['security'].confirm_register_form = \
            confirm_register_form_factory(
                app.extensions['security'].confirm_register_form, app
            )

        app.extensions['security'].login_form = login_form_factory(
            app.extensions['security'].login_form, app)

        if app.config['ACCOUNTS_USE_CELERY']:
            from invenio_accounts.tasks import send_security_email

            @state.send_mail_task
            def delay_security_email(msg):
                send_security_email.delay(msg.__dict__)

        # Register context processor
        if app.config['ACCOUNTS_JWT_DOM_TOKEN']:
            from invenio_accounts.context_processors.jwt import jwt_proccessor
            app.context_processor(jwt_proccessor)

        # Register signal receiver
        if app.config.get('ACCOUNTS_USERINFO_HEADERS'):
            request_finished.connect(set_session_info, app)

        # Set Session KV store
        session_kvstore_factory = obj_or_import_string(
            app.config['ACCOUNTS_SESSION_STORE_FACTORY'])
        session_kvstore = session_kvstore_factory(app)

        self.kvsession_extension = KVSessionExtension(
            session_kvstore, app)

        self.check_configuration_consistency(app)

        app.extensions['invenio-accounts'] = self

    def init_config(self, app):
        """Initialize configuration.

        :param app: The Flask application.
        """
        try:
            pkg_resources.get_distribution('celery')
            app.config.setdefault(
                "ACCOUNTS_USE_CELERY", not (app.debug or app.testing))
        except pkg_resources.DistributionNotFound:  # pragma: no cover
            app.config.setdefault("ACCOUNTS_USE_CELERY", False)

        # Register Invenio legacy password hashing
        register_crypt_handler(InvenioAesEncryptedEmail)

        # Change Flask defaults
        app.config.setdefault(
            'SESSION_COOKIE_SECURE',
            not app.debug
        )

        # Change Flask-Security defaults
        app.config.setdefault(
            'SECURITY_PASSWORD_SALT',
            app.config['SECRET_KEY']
        )

        # Set JWT secret key
        app.config.setdefault(
            'ACCOUNTS_JWT_SECRET_KEY',
            app.config.get(
                'ACCOUNTS_JWT_SECRET_KEY',
                app.config.get('SECRET_KEY')
            )
        )

        config_apps = ['ACCOUNTS', 'SECURITY_']
        for k in dir(config):
            if any([k.startswith(prefix) for prefix in config_apps]):
                app.config.setdefault(k, getattr(config, k))

    def _enable_session_activity(self, app):
        """Enable session activity."""
        user_logged_in.connect(login_listener, app)
        user_logged_in.connect(csrf_token_reset, app)
        user_logged_out.connect(logout_listener, app)
        user_logged_out.connect(csrf_token_reset, app)
        from .views.security import revoke_session, security
        from .views.settings import blueprint
        blueprint.route('/security/', methods=['GET'])(security)
        blueprint.route('/sessions/revoke/', methods=['POST'])(revoke_session)


class InvenioAccountsREST(InvenioAccounts):
    """Invenio-Accounts REST extension."""

    def init_app(self, app, sessionstore=None, register_blueprint=False):
        """Flask application initialization.

        :param app: The Flask application.
        :param sessionstore: store for sessions. Passed to
            ``flask-kvsession``. If ``None`` then Redis is configured.
            (Default: ``None``)
        :param register_blueprint: If ``True``, the application registers the
            blueprints. (Default: ``True``)
        """
        # Register the Flask-Security blueprint for the email templates
        if not register_blueprint:
            security_bp = Blueprint(
                'security_email_templates',  # name differently to avoid misuse
                'flask_security.core', template_folder='templates')
            security_rest_overrides = Blueprint(
                'security_email_overrides',  # overrides
                __name__, template_folder='templates')
            app.register_blueprint(security_bp)
            app.register_blueprint(security_rest_overrides)

        super(InvenioAccountsREST, self).init_app(
            app, sessionstore=sessionstore,
            register_blueprint=register_blueprint,
        )
        app.config['ACCOUNTS_CONFIRM_EMAIL_ENDPOINT'] = \
            app.config['ACCOUNTS_REST_CONFIRM_EMAIL_ENDPOINT']
        app.config['ACCOUNTS_RESET_PASSWORD_ENDPOINT'] = \
            app.config['ACCOUNTS_REST_RESET_PASSWORD_ENDPOINT']

        if app.config.get("ACCOUNTS_REGISTER_UNAUTHORIZED_CALLBACK", True):
            def _unauthorized_callback():
                """Callback to abort when user is unauthorized."""
                abort(401)
            app.login_manager.unauthorized_handler(_unauthorized_callback)


class InvenioAccountsUI(InvenioAccounts):
    """Invenio-Accounts UI extension."""

    def init_app(self, app, sessionstore=None, register_blueprint=True):
        """Flask application initialization.

        :param app: The Flask application.
        :param sessionstore: store for sessions. Passed to
            ``flask-kvsession``. If ``None`` then Redis is configured.
            (Default: ``None``)
        :param register_blueprint: If ``True``, the application registers the
            blueprints. (Default: ``True``)
        """
        self.make_session_permanent(app)
        return super(InvenioAccountsUI, self).init_app(
            app, sessionstore=sessionstore,
            register_blueprint=register_blueprint
        )

    def make_session_permanent(self, app):
        """Make session permanent by default.

        Set `PERMANENT_SESSION_LIFETIME` to specify time-to-live
        """
        @app.before_request
        def make_session_permanent():
            session.permanent = True



#
# This file is part of Invenio.
# Copyright (C) 2016-2022 CERN.
# Copyright (C)      2022 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Separate login info from user table."""

import sqlalchemy as sa
from alembic import op
from sqlalchemy_utils.types.ip_address import IPAddressType

# revision identifiers, used by Alembic.
revision = "dfbdf43a3e96"
down_revision = "999dcbd19ace"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    # create the new login information table
    op.create_table(
        "accounts_user_login_information",
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("current_login_at", sa.DateTime(), nullable=True),
        sa.Column("current_login_ip", IPAddressType(), nullable=True),
        sa.Column("last_login_at", sa.DateTime(), nullable=True),
        sa.Column("last_login_ip", IPAddressType(), nullable=True),
        sa.Column("login_count", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["accounts_user.id"],
            name="fk_accounts_login_information_user_id",
        ),
        sa.PrimaryKeyConstraint(
            "user_id", name=op.f("pk_accounts_user_login_information")
        ),
    )

    # move information from the user table into the new table
    op.execute(
        "INSERT INTO accounts_user_login_information "
        "(user_id, current_login_at, current_login_ip, last_login_at, last_login_ip, login_count) "  # noqa
        "SELECT id, current_login_at, current_login_ip, last_login_at, last_login_ip, login_count "  # noqa
        "FROM accounts_user;"
    )

    # remove columns from the user table
    op.drop_column("accounts_user", "login_count")
    op.drop_column("accounts_user", "last_login_at")
    op.drop_column("accounts_user", "current_login_ip")
    op.drop_column("accounts_user", "last_login_ip")
    op.drop_column("accounts_user", "current_login_at")


def downgrade():
    """Downgrade database."""
    # create the old columns on the user table
    op.add_column(
        "accounts_user",
        sa.Column("current_login_at", sa.DateTime(), nullable=True),
    )
    op.add_column(
        "accounts_user",
        sa.Column("current_login_ip", IPAddressType(), nullable=True),
    )
    op.add_column(
        "accounts_user",
        sa.Column("last_login_at", sa.DateTime(), nullable=True),
    )
    op.add_column(
        "accounts_user",
        sa.Column("last_login_ip", IPAddressType(), nullable=True),
    )
    op.add_column(
        "accounts_user",
        sa.Column("login_count", sa.Integer(), nullable=True),
    )

    # move the data from the new table to the old columns
    # (at least with postgres)
    dialect_name = op._proxy.migration_context.dialect.name
    if dialect_name == "postgresql":
        op.execute(
            "UPDATE accounts_user SET "
            " current_login_at = li.current_login_at, "
            " current_login_ip = li.current_login_ip, "
            " last_login_at = li.last_login_at, "
            " last_login_ip = li.last_login_ip, "
            " login_count = li.login_count "
            "FROM accounts_user_login_information AS li "
            "WHERE accounts_user.id = li.user_id;"
        )

    # drop the new table
    op.drop_table("accounts_user_login_information")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Database models for accounts."""

from datetime import datetime

from flask import current_app, session
from flask_security import RoleMixin, UserMixin
from invenio_db import db
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import validates
from sqlalchemy_utils import IPAddressType, Timestamp

from .errors import AlreadyLinkedError

userrole = db.Table(
    'accounts_userrole',
    db.Column('user_id', db.Integer(), db.ForeignKey(
        'accounts_user.id', name='fk_accounts_userrole_user_id')),
    db.Column('role_id', db.Integer(), db.ForeignKey(
        'accounts_role.id', name='fk_accounts_userrole_role_id')),
)
"""Relationship between users and roles."""


class Role(db.Model, Timestamp, RoleMixin):
    """Role data model."""

    __tablename__ = "accounts_role"

    id = db.Column(db.Integer(), primary_key=True)

    name = db.Column(db.String(80), unique=True)
    """Role name."""

    description = db.Column(db.String(255))
    """Role description."""

    # Enables SQLAlchemy version counter
    version_id = db.Column(db.Integer, nullable=False)
    """Used by SQLAlchemy for optimistic concurrency control."""

    __mapper_args__ = {
        "version_id_col": version_id
    }

    def __str__(self):
        """Return the name and description of the role."""
        return '{0.name} - {0.description}'.format(self)


class User(db.Model, Timestamp, UserMixin):
    """User data model."""

    __tablename__ = "accounts_user"

    id = db.Column(db.Integer, primary_key=True)

    email = db.Column(db.String(255), unique=True)
    """User email."""

    password = db.Column(db.String(255))
    """User password."""

    active = db.Column(db.Boolean(name='active'))
    """Flag to say if the user is active or not ."""

    confirmed_at = db.Column(db.DateTime)
    """When the user confirmed the email address."""

    roles = db.relationship('Role', secondary=userrole,
                            backref=db.backref('users', lazy='dynamic'))
    """List of the user's roles."""

    # Enables SQLAlchemy version counter
    version_id = db.Column(db.Integer, nullable=False)
    """Used by SQLAlchemy for optimistic concurrency control."""

    __mapper_args__ = {
        "version_id_col": version_id
    }

    login_info = db.relationship(
        "LoginInformation", back_populates="user", uselist=False, lazy="joined"
    )

    def _get_login_info_attr(self, attr_name):
        if self.login_info is None:
            return None
        return getattr(self.login_info, attr_name)

    def _set_login_info_attr(self, attr_name, value):
        if self.login_info is None:
            self.login_info = LoginInformation()
        setattr(self.login_info, attr_name, value)

    @property
    def current_login_at(self):
        """When user logged into the current session."""
        return self._get_login_info_attr("current_login_at")

    @property
    def current_login_ip(self):
        """Current user IP address."""
        return self._get_login_info_attr("current_login_ip")

    @property
    def last_login_at(self):
        """When the user logged-in for the last time."""
        return self._get_login_info_attr("last_login_at")

    @property
    def last_login_ip(self):
        """Last user IP address."""
        return self._get_login_info_attr("last_login_ip")

    @property
    def login_count(self):
        """Count how many times the user logged in."""
        return self._get_login_info_attr("login_count")

    @current_login_at.setter
    def current_login_at(self, value):
        return self._set_login_info_attr("current_login_at", value)

    @current_login_ip.setter
    def current_login_ip(self, value):
        return self._set_login_info_attr("current_login_ip", value)

    @last_login_at.setter
    def last_login_at(self, value):
        return self._set_login_info_attr("last_login_at", value)

    @last_login_ip.setter
    def last_login_ip(self, value):
        return self._set_login_info_attr("last_login_ip", value)

    @login_count.setter
    def login_count(self, value):
        return self._set_login_info_attr("login_count", value)

    def __str__(self):
        """Representation."""
        return 'User <id={0.id}, email={0.email}>'.format(self)


class LoginInformation(db.Model):
    """Login information for a user."""

    __tablename__ = "accounts_user_login_information"

    user_id = db.Column(
        db.Integer,
        db.ForeignKey(User.id, name='fk_accounts_login_information_user_id'),
        primary_key=True,
    )
    """ID of user to whom this information belongs."""

    user = db.relationship("User", back_populates="login_info")
    """User to whom this information belongs."""

    last_login_at = db.Column(db.DateTime)
    """When the user logged-in for the last time."""

    current_login_at = db.Column(db.DateTime)
    """When user logged into the current session."""

    last_login_ip = db.Column(IPAddressType, nullable=True)
    """Last user IP address."""

    current_login_ip = db.Column(IPAddressType, nullable=True)
    """Current user IP address."""

    login_count = db.Column(db.Integer)
    """Count how many times the user logged in."""

    @validates('last_login_ip', 'current_login_ip')
    def validate_ip(self, key, value):
        """Hack untrackable IP addresses."""
        # NOTE Flask-Security stores 'untrackable' value to IPAddressType
        #      field. This incorrect value causes ValueError on loading
        #      user object.
        if value == 'untrackable':  # pragma: no cover
            value = None
        return value


class SessionActivity(db.Model, Timestamp):
    """User Session Activity model.

    Instances of this model correspond to a session belonging to a user.
    """

    __tablename__ = "accounts_user_session_activity"

    sid_s = db.Column(db.String(255), primary_key=True)
    """Serialized Session ID. Used as the session's key in the kv-session
    store employed by `flask-kvsession`.
    Named here as it is in `flask-kvsession` to avoid confusion.
    """

    user_id = db.Column(db.Integer, db.ForeignKey(
        User.id, name='fk_accounts_session_activity_user_id'))
    """ID of user to whom this session belongs."""

    user = db.relationship(User, backref='active_sessions')

    ip = db.Column(db.String(80), nullable=True)
    """IP address."""

    country = db.Column(db.String(3), nullable=True)
    """Country name."""

    browser = db.Column(db.String(80), nullable=True)
    """User browser."""

    browser_version = db.Column(db.String(30), nullable=True)
    """Browser version."""

    os = db.Column(db.String(80), nullable=True)
    """User operative system name."""

    device = db.Column(db.String(80), nullable=True)
    """User device."""

    @classmethod
    def query_by_expired(cls):
        """Query to select all expired sessions."""
        lifetime = current_app.permanent_session_lifetime
        expired_moment = datetime.utcnow() - lifetime
        return cls.query.filter(cls.created < expired_moment)

    @classmethod
    def query_by_user(cls, user_id):
        """Query to select user sessions."""
        return cls.query.filter_by(user_id=user_id)

    @classmethod
    def is_current(cls, sid_s):
        """Check if the session is the current one."""
        return session.sid_s == sid_s


class UserIdentity(db.Model, Timestamp):
    """Represent a UserIdentity record."""

    __tablename__ = 'accounts_useridentity'

    id = db.Column(db.String(255), primary_key=True, nullable=False)
    method = db.Column(db.String(255), primary_key=True, nullable=False)
    id_user = db.Column(db.Integer(),
                        db.ForeignKey(User.id), nullable=False)

    user = db.relationship(User, backref='external_identifiers')

    __table_args__ = (
        db.Index('useridentity_id_user_method', id_user, method, unique=True),
    )

    @classmethod
    def get_user(cls, method, external_id):
        """Get the user for a given identity."""
        identity = cls.query.filter_by(
            id=external_id, method=method).one_or_none()
        if identity is not None:
            return identity.user
        return None

    @classmethod
    def create(cls, user, method, external_id):
        """Link a user to an external id.

        :param user: A :class:`invenio_accounts.models.User` instance.
        :param method: The identity source (e.g. orcid, github)
        :param method: The external identifier.
        :raises AlreadyLinkedError: Raised if already exists a link.
        """
        try:
            with db.session.begin_nested():
                db.session.add(cls(
                    id=external_id,
                    method=method,
                    id_user=user.id
                ))
        except IntegrityError:
            raise AlreadyLinkedError(
                # dict used for backward compatibility (came from oauthclient)
                user, {"id": external_id, "method": method})

    @classmethod
    def delete_by_external_id(cls, method, external_id):
        """Unlink a user from an external id."""
        with db.session.begin_nested():
            cls.query.filter_by(id=external_id, method=method).delete()

    @classmethod
    def delete_by_user(cls, method, user):
        """Unlink a user from an external id."""
        with db.session.begin_nested():
            cls.query.filter_by(id_user=user.id, method=method).delete()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""JWT context processors."""

from flask import current_app, render_template
from jinja2 import Markup

from ..proxies import current_accounts


def jwt_proccessor():
    """Context processor for jwt."""
    def jwt():
        """Context processor function to generate jwt."""
        token = current_accounts.jwt_creation_factory()
        return Markup(
            render_template(
                current_app.config['ACCOUNTS_JWT_DOM_TOKEN_TEMPLATE'],
                token=token
            )
        )

    def jwt_token():
        """Context processor function to generate jwt."""
        return current_accounts.jwt_creation_factory()

    return {
        'jwt': jwt,
        'jwt_token': jwt_token,
    }



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Exception classes."""


class JWTExtendedException(Exception):
    """Base exception for all JWT errors."""


class JWTDecodeError(JWTExtendedException):
    """Exception raised when decoding is failed."""


class JWTExpiredToken(JWTExtendedException):
    """Exception raised when JWT is expired."""


class AlreadyLinkedError(Exception):
    """Signifies that an account was already linked to another account."""

    def __init__(self, user, external_id):
        """Initialize exception."""
        self.user = user
        self.external_id = external_id



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test backend sessions."""

import datetime
import time

import flask_security
import pytest
from flask import current_app, session
from flask_security import url_for_security
from invenio_db import db
from simplekv.memory.redisstore import RedisStore
from werkzeug.local import LocalProxy

from invenio_accounts import testutils
from invenio_accounts.models import SessionActivity
from invenio_accounts.sessions import delete_session

_sessionstore = LocalProxy(lambda: current_app.
                           extensions['invenio-accounts'].sessionstore)
_datastore = LocalProxy(lambda: current_app.extensions['security'].datastore)


def test_login_listener(app):
    """Test login listener."""
    with app.app_context():
        with app.test_client() as client:
            user = testutils.create_test_user('test@example.org')
            # The SessionActivity table is initially empty
            query = db.session.query(SessionActivity)
            assert query.count() == 0

            testutils.login_user_via_view(client, user.email,
                                          user.password_plaintext)
            assert testutils.client_authenticated(client)
            # After logging in, a SessionActivity has been created
            # corresponding to the user's session.
            query = db.session.query(SessionActivity)
            assert query.count() == 1

            session_entry = query.first()
            assert session_entry.user_id == user.id
            assert session_entry.sid_s == session.sid_s


def test_repeated_login_session_population(app):
    """Verify session population on repeated login."""
    with app.app_context():
        user = testutils.create_test_user('test@example.org')
        query = db.session.query(SessionActivity)
        assert query.count() == len(app.kvsession_store.keys())

        with app.test_client() as client:
            # After logging in, there should be one session in the kv-store and
            # one SessionActivity
            testutils.login_user_via_view(client, user=user)
            assert testutils.client_authenticated(client)
            query = db.session.query(SessionActivity)
            assert query.count() == 1
            assert len(app.kvsession_store.keys()) == 1
            first_login_session_id = app.kvsession_store.keys()[0]

            # SessionActivity is deleted upon logout
            client.get(flask_security.url_for_security('logout'))
            query = db.session.query(SessionActivity)
            assert query.count() == 0
            # Session id changes after logout
            assert len(app.kvsession_store.keys()) == 1
            assert first_login_session_id != app.kvsession_store.keys()[0]

            # After logging out and back in, the should be one
            # SessionActivity entry.
            testutils.login_user_via_view(client, user=user)
            query = db.session.query(SessionActivity)
            assert query.count() == 1
            assert len(app.kvsession_store.keys()) == 1
            assert first_login_session_id != app.kvsession_store.keys()[0]


def test_login_multiple_clients_single_user_session_population(app):
    """Test session population/creation from multiple clients for same user."""
    with app.app_context():
        user = testutils.create_test_user('test@example.org')

        client_count = 3
        clients = [app.test_client() for _ in range(client_count)]
        sid_s_list = []
        for c in clients:
            with c as client:
                testutils.login_user_via_view(client, user=user)
                assert testutils.client_authenticated(client)
                sid_s_list.append(session.sid_s)

        # There is now `client_count` existing sessions and SessionActivity
        # entries
        assert len(app.kvsession_store.keys()) == client_count
        query = db.session.query(SessionActivity)
        assert query.count() == client_count
        assert len(user.active_sessions) == client_count


def test_sessionstore_default_ttl_secs(app):
    """Test the `default_ttl_secs` field for simplekv sessionstore.

    See http://simplekv.readthedocs.io/index.html#simplekv.TimeToLiveMixin.
    """
    if type(app.kvsession_store) is not RedisStore:
        pytest.skip('TTL support needed, this test requires Redis.')

    ttl_seconds = 3
    ttl_delta = datetime.timedelta(0, ttl_seconds)

    sessionstore = app.kvsession_store
    sessionstore.default_ttl_secs = ttl_seconds

    # Verify that the backend supports ttl
    assert sessionstore.ttl_support

    app.kvsession_store = sessionstore

    with app.app_context():
        user = testutils.create_test_user('test@example.org')

        with app.test_client() as client:
            testutils.login_user_via_view(client, user=user)
            sid = testutils.unserialize_session(session.sid_s)
            while not sid.has_expired(ttl_delta):
                pass
            # When we get here the session should have expired.
            # But the client is still authenticated.
            assert testutils.client_authenticated(client)
            # Why? Because `flask_kvsession` doesn't care about
            # the sessionstore's `default_ttl_seconds`. It uses
            # the `PERMANENT_SESSION_LIFETIME` from the app's config.


def test_session_ttl(app):
    """Test actual/working session expiration/TTL settings."""
    if type(app.kvsession_store) is not RedisStore:
        pytest.skip('TTL support needed, this test requires Redis.')

    ttl_seconds = 3
    # Set ttl to "0 days, 1 seconds"
    ttl_delta = datetime.timedelta(0, ttl_seconds)

    assert app.kvsession_store.ttl_support

    # _THIS_ is what flask_kvsession uses to determine default ttl
    # sets default ttl to `ttl_seconds` seconds
    app.config['PERMANENT_SESSION_LIFETIME'] = ttl_delta
    assert app.permanent_session_lifetime.total_seconds() == ttl_seconds

    with app.app_context():
        user = testutils.create_test_user('test@example.org')

        with app.test_client() as client:
            testutils.login_user_via_view(client, user=user)
            assert len(app.kvsession_store.keys()) == 1

            sid = testutils.unserialize_session(session.sid_s)
            time.sleep(ttl_seconds + 1)

            assert sid.has_expired(ttl_delta)
            assert not testutils.client_authenticated(client)

            # Expired sessions are automagically removed from the sessionstore
            # Although not _instantly_.
            while len(app.kvsession_store.keys()) > 0:
                pass
            assert len(app.kvsession_store.keys()) == 0


def test_repeated_login_session_expiration(app):
    """Test repeated session login.

    Test that a new session (with a different sid_s) is created when logging
    in again after a previous session has expired.
    """
    ttl_seconds = 3
    ttl_delta = datetime.timedelta(0, ttl_seconds)
    app.config['PERMANENT_SESSION_LIFETIME'] = ttl_delta

    with app.app_context():
        user = testutils.create_test_user('test@example.org')
        with app.test_client() as client:
            testutils.login_user_via_view(client, user=user)
            first_sid_s = session.sid_s
            time.sleep(ttl_seconds + 1)
            assert not testutils.client_authenticated(client)

            app.config['PERMANENT_SESSION_LIFETIME'] = datetime.timedelta(
                0, 10000)
            testutils.login_user_via_view(client, user=user)
            second_sid_s = session.sid_s

            assert not first_sid_s == second_sid_s


def test_session_deletion(app):
    """Test if user is not authenticated when session is deleted."""
    with app.app_context():
        user = testutils.create_test_user('test@example.org')

        with app.test_client() as client:
            testutils.login_user_via_view(client, user=user)
            assert testutils.client_authenticated(client)
            assert len(user.active_sessions) == 1
            saved_sid_s = session.sid_s

            delete_session(saved_sid_s)
            db.session.commit()
            # The user now has no active sessions
            assert len(app.kvsession_store.keys()) == 0
            assert len(user.active_sessions) == 0
            query = db.session.query(SessionActivity)
            assert query.count() == 0

            # After deleting the session, the client is not authenticated
            assert not testutils.client_authenticated(client)

            # A new session is created in the kv-sessionstore, but its
            # sid_s is different and the user is not authenticated with it.
            assert len(app.kvsession_store.keys()) == 1
            assert not session.sid_s == saved_sid_s
            assert not testutils.client_authenticated(client)


def test_deactivate_user(app):
    """Test deactivation of users."""
    with app.app_context():
        user_bob = testutils.create_test_user(email='bob@bobmail.bob',
                                              password='123',
                                              active=True)
        with app.test_client() as client:
            assert user_bob.active
            assert not testutils.client_authenticated(client)
            testutils.login_user_via_view(client, user_bob.email,
                                          user_bob.password_plaintext)
            assert testutils.client_authenticated(client)
            # Now we deactivate Bob.
            # `deactivate_user` returns True if a change was made.
            _datastore.deactivate_user(user_bob)
            db.session.commit()
            assert not testutils.client_authenticated(client)


def test_session_extra_info_on_login(app, users):
    """Test session extra info on login."""
    ua = 'Mozilla/5.0 (X11; Linux x86_64; rv:43.0) Gecko/20100101 Firefox/43.0'

    with app.test_client() as client:
        res = client.post(
            url_for_security('login'),
            data={
                'email': users[0]['email'],
                'password': users[0]['password']
            },
            environ_base={'REMOTE_ADDR': '188.184.9.234'},
            headers={'User-Agent': ua}
        )
        assert res.status_code == 302
        # check if session extra info is there
        [session] = SessionActivity.query.all()
        assert session.browser == 'Firefox'
        assert session.browser_version == '43'
        assert session.device == 'Other'
        assert session.os == 'Linux'
        assert session.country == 'CH'
        assert session.ip == '188.184.9.234'


def test_session_ip_no_country(app, users):
    """Test session with an IP without country information."""
    ua = 'Mozilla/5.0 (X11; Linux x86_64; rv:43.0) Gecko/20100101 Firefox/43.0'

    with app.test_client() as client:
        res = client.post(
            url_for_security('login'),
            data={
                'email': users[0]['email'],
                'password': users[0]['password']
            },
            environ_base={'REMOTE_ADDR': '139.191.247.1'},
            headers={'User-Agent': ua}
        )
        assert res.status_code == 302
        [session] = SessionActivity.query.all()
        assert session.country is None
        assert session.ip == '139.191.247.1'



#
# This file is part of Invenio.
# Copyright (C) 2016-2022 CERN.
# Copyright (C)      2022 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Add versioning information to models."""

from alembic import op
from sqlalchemy import Column, DateTime, Integer

# revision identifiers, used by Alembic.
revision = "999dcbd19ace"
down_revision = "e12419831262"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    role_tbl = "accounts_role"
    user_tbl = "accounts_user"

    # create the new columns as nullable fields
    op.add_column(role_tbl, Column("created", DateTime(), nullable=True))
    op.add_column(role_tbl, Column("updated", DateTime(), nullable=True))
    op.add_column(role_tbl, Column("version_id", Integer(), nullable=True))
    op.add_column(user_tbl, Column("created", DateTime(), nullable=True))
    op.add_column(user_tbl, Column("updated", DateTime(), nullable=True))
    op.add_column(user_tbl, Column("version_id", Integer(), nullable=True))

    # populate the columns of existing rows
    op.execute("UPDATE accounts_role SET created = CURRENT_TIMESTAMP, updated = CURRENT_TIMESTAMP, version_id = 1;")  # noqa
    op.execute("UPDATE accounts_user SET created = CURRENT_TIMESTAMP, updated = CURRENT_TIMESTAMP, version_id = 1;")  # noqa

    # make the columns not nullable
    dt_args = {"existing_type": DateTime(), "nullable": False}
    int_args = {"existing_type": Integer(), "nullable": False}
    op.alter_column(role_tbl, "created", **dt_args)
    op.alter_column(role_tbl, "updated", **dt_args)
    op.alter_column(role_tbl, "version_id", **int_args)
    op.alter_column(user_tbl, "created", **dt_args)
    op.alter_column(user_tbl, "updated", **dt_args)
    op.alter_column(user_tbl, "version_id", **int_args)


def downgrade():
    """Downgrade database."""
    op.drop_column("accounts_user", "version_id")
    op.drop_column("accounts_user", "updated")
    op.drop_column("accounts_user", "created")

    op.drop_column("accounts_role", "version_id")
    op.drop_column("accounts_role", "updated")
    op.drop_column("accounts_role", "created")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio user management and authentication."""

from flask import Blueprint, current_app
from flask_babelex import lazy_gettext as _
from flask_breadcrumbs import register_breadcrumb
from flask_menu import current_menu
from invenio_theme.proxies import current_theme_icons

blueprint = Blueprint(
    'invenio_accounts',
    __name__,
    url_prefix='/account/settings',
    template_folder='../templates',
    static_folder='static',
)


@blueprint.record_once
def post_ext_init(state):
    """."""
    app = state.app

    app.config.setdefault(
        "ACCOUNTS_SITENAME",
        app.config.get("THEME_SITENAME", "Invenio"))
    app.config.setdefault(
        "ACCOUNTS_BASE_TEMPLATE",
        app.config.get("BASE_TEMPLATE",
                       "invenio_accounts/base.html"))
    app.config.setdefault(
        "ACCOUNTS_COVER_TEMPLATE",
        app.config.get("COVER_TEMPLATE",
                       "invenio_accounts/base_cover.html"))
    app.config.setdefault(
        "ACCOUNTS_SETTINGS_TEMPLATE",
        app.config.get("SETTINGS_TEMPLATE",
                       "invenio_accounts/settings/base.html"))


@blueprint.before_app_first_request
def init_menu():
    """Initialize menu before first request."""
    # Register root breadcrumbs
    item = current_menu.submenu('breadcrumbs.settings')
    item.register('invenio_userprofiles.profile', _('Account'))

    # - Register menu
    # - Change password
    if current_app.config.get('SECURITY_CHANGEABLE', True):
        view_name = '{}.change_password'.format(
            current_app.config['SECURITY_BLUEPRINT_NAME'])

        item = current_menu.submenu('settings.change_password')
        item.register(
            view_name,
            # NOTE: Menu item text (icon replaced by a key icon).
            _('%(icon)s Change password',
                icon=f'<i class="{current_theme_icons.key}"></i>'),
            order=1)

        # Breadcrumb for change password
        #
        # The breadcrumbs works by decorating the view functions with a
        # __breadcrumb__ field. Since the change password view is defined in
        # Flask-Security, we need to this hack to in order to decorate the view
        # function with the __breadcrumb__ field.
        decorator = register_breadcrumb(
            current_app,
            'breadcrumbs.settings.change_password',
            _('Change password')
        )
        current_app.view_functions[view_name] = decorator(
            current_app.view_functions[view_name])


@blueprint.before_app_first_request
def check_security_settings():
    """Warn if session cookie is not secure in production."""
    in_production = not (current_app.debug or current_app.testing)
    secure = current_app.config.get('SESSION_COOKIE_SECURE')
    if in_production and not secure:
        current_app.logger.warning(
            "SESSION_COOKIE_SECURE setting must be set to True to prevent the "
            "session cookie from being leaked over an insecure channel."
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
# Copyright (C)      2021 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Default configuration for ACCOUNTS."""

from datetime import timedelta

from .views import login

ACCOUNTS = True
"""Tells if the templates should use the accounts module.

If False, you won't be able to login via the web UI.
"""

ACCOUNTS_RETENTION_PERIOD = timedelta(days=30)

ACCOUNTS_SESSION_STORE_FACTORY = \
    'invenio_accounts.sessions:default_session_store_factory'
"""Import path or function of factory used to generate the session store
object.

When  ``ACCOUNTS_SESSION_REDIS_URL`` will use redis as cache system otherwise
otherwise it will use the in-memory backend :class:`simplekv.memory.DictStore`.
"""

ACCOUNTS_SESSION_REDIS_URL = None
"""Redis URL used by the module as a cache system for sessions."""

ACCOUNTS_REGISTER_BLUEPRINT = None
"""Register the Security blueprint or not.

It can be used to override the ``register_blueprint`` option.

.. note:: If the value is ``None``, then the blueprint is not registered.
"""

ACCOUNTS_USE_CELERY = None
"""Tells if the module should use Celery or not.

By default, it uses Celery if it can find it.
"""

ACCOUNTS_SESSION_ACTIVITY_ENABLED = True
"""Enable session activity tracking."""

ACCOUNTS_SETTINGS_SECURITY_TEMPLATE = 'invenio_accounts/settings/security.html'
"""Template for the account security page."""

ACCOUNTS_CONFIRM_EMAIL_ENDPOINT = None
"""Value to be used for the confirmation email link in the UI application."""

ACCOUNTS_REST_CONFIRM_EMAIL_ENDPOINT = '/confirm/{token}'
"""Value to be used for the confirmation email link in the API application.

Can be a Flask endpoint (e.g. "invenio_accounts_rest_auth.confirm_email"), or
a URL part (e.g. "https://ui.example.com/confirm-email", "/confirm-email").

This will be used to build an absolute URL, thus if e.g. a hostname isn't
included, the one from the current request's context will be used.
"""

ACCOUNTS_RESET_PASSWORD_ENDPOINT = None
"""Value to be used for the confirmation email link in the UI application."""

ACCOUNTS_REST_RESET_PASSWORD_ENDPOINT = '/lost-password/{token}'
"""Value to be used for the reset password link in the API application.

Can be a Flask endpoint (e.g. "invenio_accounts_rest_auth.reset_password"), or
a URL part (e.g. "https://ui.example.com/reset-password", "/reset-password").

This will be used to build an absolute URL, thus if e.g. a hostname isn't
included, the one from the current request's context will be used.
"""

ACCOUNTS_REST_AUTH_VIEWS = {
    "login": "invenio_accounts.views.rest:LoginView",
    "logout": "invenio_accounts.views.rest:LogoutView",
    "user_info": "invenio_accounts.views.rest:UserInfoView",
    "register": "invenio_accounts.views.rest:RegisterView",
    "forgot_password": "invenio_accounts.views.rest:ForgotPasswordView",
    "reset_password": "invenio_accounts.views.rest:ResetPasswordView",
    "change_password": "invenio_accounts.views.rest:ChangePasswordView",
    "send_confirmation":
        "invenio_accounts.views.rest:SendConfirmationEmailView",
    "confirm_email": "invenio_accounts.views.rest:ConfirmEmailView",
    "sessions_list": "invenio_accounts.views.rest:SessionsListView",
    "sessions_item": "invenio_accounts.views.rest:SessionsItemView"
}
"""List of REST API authentication views."""

# Change Flask-Security defaults
SECURITY_PASSWORD_HASH = 'pbkdf2_sha512'
"""Default password hashing algorithm for new passwords."""

SECURITY_PASSWORD_SCHEMES = ['pbkdf2_sha512', 'invenio_aes_encrypted_email']
"""Supported password hashing algorithms (for passwords already stored).

You should include both the default, supported and any deprecated schemes.
"""

SECURITY_DEPRECATED_PASSWORD_SCHEMES = ['invenio_aes_encrypted_email']
"""Deprecated password hashing algorithms.

Password hashes in a deprecated scheme are automatically migrated to the
new default algorithm the next time the user login.
"""

SECURITY_PASSWORD_SINGLE_HASH = ['invenio_aes_encrypted_email']
"""Password hashing algorithms requiring single hasing only."""

SECURITY_DEFAULT_REMEMBER_ME = False
""""Remember me" default value in login form.

This is only the default value in the login form. A user can always choose to
change it when they login.
"""

SECURITY_CHANGEABLE = True
"""Allow password change by users."""

SECURITY_CONFIRMABLE = True
"""Allow user to confirm their email address."""

SECURITY_RECOVERABLE = True
"""Allow password recovery by users."""

SECURITY_REGISTERABLE = True
"""Allow users to register."""

SECURITY_CONFIRM_EMAIL_WITHIN = '30 minutes'
"""Amount of time the email confirmation link is active.

Note, since the confirmation link will also login the associated user we expire
the link fast.
"""

SECURITY_RESET_PASSWORD_WITHIN = '30 minutes'
"""Amount of time the password reset link is active.

Note, since the confirmation link will also login the associated user we expire
the link fast.
"""

SECURITY_TRACKABLE = True
"""Enable user tracking on login."""

SECURITY_LOGIN_WITHOUT_CONFIRMATION = True
"""Allow users to login without first confirming their email address."""

SECURITY_PASSWORD_SALT = None
"""Salt for storing passwords."""

# Change default templates
SECURITY_FORGOT_PASSWORD_TEMPLATE = 'invenio_accounts/forgot_password.html'
"""Default template for password recovery (asking for email)."""

SECURITY_LOGIN_USER_TEMPLATE = 'invenio_accounts/login_user.html'
"""Default template for login."""

SECURITY_REGISTER_USER_TEMPLATE = 'invenio_accounts/register_user.html'
"""Default template for user registration."""

SECURITY_RESET_PASSWORD_TEMPLATE = 'invenio_accounts/reset_password.html'
"""Default template for password recovery (reset of the password)."""

SECURITY_CHANGE_PASSWORD_TEMPLATE = 'invenio_accounts/change_password.html'
"""Default template for change password."""

SECURITY_SEND_CONFIRMATION_TEMPLATE = \
    'invenio_accounts/send_confirmation.html'
"""Default template for email confirmation."""

SECURITY_SEND_LOGIN_TEMPLATE = 'invenio_accounts/send_login.html'
"""Default template for email confirmation."""

SECURITY_REGISTER_URL = '/signup/'
"""URL endpoint for user registation."""

SECURITY_RESET_URL = '/lost-password/'
"""URL endpoint for password recovery."""

SECURITY_LOGIN_URL = '/login/'
"""URL endpoint for login."""

SECURITY_LOGOUT_URL = '/logout/'
"""URL endpoint for logout."""

SECURITY_CHANGE_URL = '/account/settings/password/'
"""URL endpoint for password change."""

SECURITY_MSG_LOCAL_LOGIN_DISABLED = ("Local login is disabled.", "error")
"""The error to be displayed in REST login when local login is disabled."""

SECURITY_MSG_REGISTRATION_DISABLED = ("Registration is disabled.", "error")
"""The error to be displayed in REST registration when it is disabled."""

SECURITY_MSG_PASSWORD_CHANGE_DISABLED = (
    "Password change is disabled.", "error"
)
"""The error to be displayed in REST password change when it is disabled."""

SECURITY_MSG_PASSWORD_RECOVERY_DISABLED = (
    "Password recovery is disabled.", "error"
)
"""The error to be displayed in REST password recovery when it is disabled."""

SECURITY_MSG_PASSWORD_RESET_DISABLED = ("Password reset is disabled.", "error")
"""The error to be displayed in REST password reset when it is disabled."""

REMEMBER_COOKIE_DURATION = timedelta(days=90)
"""Remember me cookie life time changed to 90 days instead of 365 days."""

# JWT related config
ACCOUNTS_JWT_ENABLE = True
"""Enable JWT support.

.. note::

    More details about `JWT <https://jwt.io>`_
"""

ACCOUNTS_JWT_DOM_TOKEN = True
"""Register JWT context processor.

.. code-block:: html

    {% if current_user.is_authenticated %}
        {{ jwt() }}
    {% endif %}

This will generate a ``hidden`` field as follows:

.. code-block:: html

    <input type="hidden" name="authorized_token" value="xxx">

On your API call you can use it with simple javascript, an example using
``jQuery`` is the following:

.. code-block:: javascript

    $.ajax({
        url: '/example',
        method: 'POST',
        beforeSend: function(request) {
            request.setRequestHeader(
                'Authorization',
                'Bearer ' + $('[name=authorized_token]').val()
            );
        },
    });
"""

ACCOUNTS_JWT_DOM_TOKEN_TEMPLATE = 'invenio_accounts/jwt.html'
"""Template for the context processor."""

ACCOUNTS_JWT_SECRET_KEY = None
"""Secret key for JWT.

.. note::

    If is set to ``None`` it will use the ``SECRET_KEY``.
"""

ACCOUNTS_JWT_EXPIRATION_DELTA = timedelta(days=1)
"""Token expiration period for JWT."""

ACCOUNTS_JWT_ALOGORITHM = 'HS256'
"""Set JWT encryption alogirthm.

.. note::

   `Available aglorithms
   <https://pyjwt.readthedocs.io/en/latest/algorithms.html>`_
"""

ACCOUNTS_JWT_DECODE_FACTORY = 'invenio_accounts.utils:jwt_decode_token'
"""Import path of factory used to decode JWT."""

ACCOUNTS_JWT_CREATION_FACTORY = 'invenio_accounts.utils:jwt_create_token'
"""Import path of factory used to generate JWT."""

RECAPTCHA_PUBLIC_KEY = None
"""reCAPTCHA public key."""

RECAPTCHA_PRIVATE_KEY = None
"""reCAPTCHA private key."""

ACCOUNTS_USERINFO_HEADERS = False
"""If True, add X-Session-ID and X-User-ID to the HTTP response."""

ACCOUNTS_LOGIN_VIEW_FUNCTION = login
"""The view function to use for the login endpoint.

This can be either an import string, or the view function itself.
If set to None, the default login view function from Flask-Security will be
left as is.
"""

ACCOUNTS_LOCAL_LOGIN_ENABLED = True
"""Whether or not login with local account credentials should be enabled."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration."""

import multiprocessing
import os
import shutil
import tempfile
import time

import pytest
from flask import Flask
from flask_babelex import Babel
from flask_mail import Mail
from flask_menu import Menu
from invenio_db import InvenioDB, db
from selenium import webdriver
from sqlalchemy_utils.functions import create_database, database_exists, \
    drop_database

from invenio_accounts import InvenioAccounts
from invenio_accounts.views.settings import blueprint


@pytest.fixture(scope='session')
def app(request):
    """Flask application fixture for E2E/integration/selenium tests.

    Overrides the `app` fixture found in `../conftest.py`. Tests/files in this
    folder and subfolders will see this variant of the `app` fixture.
    """
    instance_path = tempfile.mkdtemp()
    app = Flask('testapp', instance_path=instance_path)
    app.config.update(
        ACCOUNTS_USE_CELERY=False,
        CELERY_ALWAYS_EAGER=True,
        CELERY_CACHE_BACKEND="memory",
        CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
        CELERY_RESULT_BACKEND="cache",
        LOGIN_DISABLED=False,
        MAIL_SUPPRESS_SEND=True,
        SECRET_KEY="CHANGE_ME",
        SECURITY_PASSWORD_SALT="CHANGE_ME_ALSO",
        SQLALCHEMY_DATABASE_URI=os.environ.get(
            'SQLALCHEMY_DATABASE_URI', 'sqlite:///test.db'),
        TESTING=True,
    )
    Menu(app)
    Babel(app)
    Mail(app)
    InvenioDB(app)
    InvenioAccounts(app)
    app.register_blueprint(blueprint)

    with app.app_context():
        if not database_exists(str(db.engine.url)):
            create_database(str(db.engine.url))
        db.create_all()

    def teardown():
        with app.app_context():
            drop_database(str(db.engine.url))
        shutil.rmtree(instance_path)

    request.addfinalizer(teardown)
    return app


def pytest_generate_tests(metafunc):
    """Override pytest's default test collection function.

    For each test in this directory which uses the `env_browser` fixture, said
    test is called once for each value found in the `E2E_WEBDRIVER_BROWSERS`
    environment variable.
    """
    browsers = os.environ.get('E2E_WEBDRIVER_BROWSERS', '').split()

    if not browsers:
        pytest.skip('E2E_WEBDRIVER_BROWSERS not set, '
                    'end-to-end tests skipped.')

    if 'env_browser' in metafunc.fixturenames:
        # In Python 2.7 the fallback kwarg of os.environ.get is `failobj`,
        # in 3.x it's `default`.
        metafunc.parametrize('env_browser', browsers, indirect=True)


@pytest.fixture()
def env_browser(request):
    """Create a webdriver instance of the browser specified by request.

    The default browser is Firefox.  The webdriver instance is killed after the
    number of seconds specified by the ``E2E_WEBDRIVER_TIMEOUT`` variable or
    defaults to 300 (five minutes).
    """
    timeout = int(os.environ.get('E2E_WEBDRIVER_TIMEOUT', 300))

    def wait_kill():
        time.sleep(timeout)
        browser.quit()

    def finalizer():
        browser.quit()
        timeout_process.terminate()

    timeout_process = multiprocessing.Process(target=wait_kill)

    # Create instance of webdriver.`request.param`()
    browser = getattr(webdriver, request.param)()
    # Add finalizer to quit the webdriver instance
    request.addfinalizer(finalizer)

    timeout_process.start()
    return browser



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio user management and authentication.

Adminstration interface
-----------------------
You can view and manage users and roles via the administration interface. Below
is a screenshot from the user creation:

.. image:: admin.png

Command-line interface
----------------------
Users and roles can be created via the CLI. Below is a simple example of
creating a user, a role and assining the user to the role:

.. code-block:: console

    $ flask users create --active info@inveniosoftware.org
    $ flask roles create admins
    $ flask roles add info@inveniosoftware.org admins

You can also e.g. deactive users:

.. code-block:: console

    $ flask users deactivate info@inveniosoftware.org
"""

from .ext import InvenioAccounts, InvenioAccountsREST, InvenioAccountsUI
from .proxies import current_accounts
from .version import __version__

__all__ = (
    '__version__',
    'current_accounts',
    'InvenioAccounts',
    'InvenioAccountsUI',
    'InvenioAccountsREST',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test forms."""

from flask_security.forms import ConfirmRegisterForm, RegisterForm

from invenio_accounts.forms import confirm_register_form_factory, \
    register_form_factory


def test_confirm_register_form_factory(app):
    """Test factory."""
    form = confirm_register_form_factory(ConfirmRegisterForm, app)
    assert not hasattr(form, 'recaptcha')
    app.config.update(dict(
        RECAPTCHA_PUBLIC_KEY='test',
        RECAPTCHA_PRIVATE_KEY='test',
    ))
    form = confirm_register_form_factory(ConfirmRegisterForm, app)
    assert hasattr(form, 'recaptcha')


def test_register_form_factory(app):
    """Test factory."""
    form = register_form_factory(RegisterForm, app)
    assert not hasattr(form, 'recaptcha')
    app.config.update(dict(
        RECAPTCHA_PUBLIC_KEY='test',
        RECAPTCHA_PRIVATE_KEY='test',
    ))
    form = register_form_factory(RegisterForm, app)
    assert hasattr(form, 'recaptcha')



#
# This file is part of Invenio.
# Copyright (C) 2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create UserIdentity table."""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '62efc52773d4'
down_revision = 'e12419831262'
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.create_table(
        'accounts_useridentity',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column('id', sa.String(length=255), nullable=False),
        sa.Column('method', sa.String(length=255), nullable=False),
        sa.Column('id_user', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ['id_user'],
            ['accounts_user.id'],
            name=op.f('fk_accounts_useridentity_id_user_accounts_user')
        ),
        sa.PrimaryKeyConstraint(
            'id',
            'method',
            name=op.f('pk_accounts_useridentity')
        )
    )
    op.create_index(
        'useridentity_id_user_method',
        'accounts_useridentity',
        ['id_user', 'method'],
        unique=True
    )


def downgrade():
    """Downgrade database."""
    op.drop_index(
        'useridentity_id_user_method',
        table_name='accounts_useridentity'
    )
    op.drop_table('accounts_useridentity')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Additional non-userprofile fields used during registration.

Currently supported: recaptcha
"""
from flask import request
from flask_babelex import gettext as _
from flask_security.forms import NextFormMixin
from flask_wtf import FlaskForm, Recaptcha, RecaptchaField
from wtforms import FormField, HiddenField


class RegistrationFormRecaptcha(FlaskForm):
    """Form for editing user profile."""

    recaptcha = RecaptchaField(validators=[
        Recaptcha(message=_("Please complete the reCAPTCHA."))])


class RevokeForm(FlaskForm):
    """Form for revoking a session."""

    sid_s = HiddenField()


def confirm_register_form_factory(Form, app):
    """Return confirmation for extended registration form."""
    class ConfirmRegisterForm(Form, NextFormMixin):

        def __init__(self, *args, **kwargs):
            """Adds next parameter from URL."""
            super(ConfirmRegisterForm, self).__init__(*args, **kwargs)
            if not self.next.data:
                self.next.data = request.args.get('next', '')

    if app.config.get('RECAPTCHA_PUBLIC_KEY') and \
            app.config.get('RECAPTCHA_PRIVATE_KEY'):
        class ConfirmRegisterWithCaptchaForm(ConfirmRegisterForm):
            recaptcha = FormField(RegistrationFormRecaptcha, separator='.')

        return ConfirmRegisterWithCaptchaForm

    return ConfirmRegisterForm


def register_form_factory(Form, app):
    """Return extended registration form."""
    if app.config.get('RECAPTCHA_PUBLIC_KEY') and \
            app.config.get('RECAPTCHA_PRIVATE_KEY'):
        class RegisterForm(Form):
            recaptcha = FormField(RegistrationFormRecaptcha, separator='.')

        return RegisterForm

    return Form


def login_form_factory(Form, app):
    """Return extended login form."""
    class LoginForm(Form):
        pass

    return LoginForm



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Add new columns on SessionActivity."""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'e12419831262'
down_revision = '9848d0149abd'
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    with op.batch_alter_table('accounts_user_session_activity') as batch_op:
        batch_op.add_column(sa.Column('browser', sa.String(80), nullable=True))
        batch_op.add_column(
            sa.Column('browser_version', sa.String(30), nullable=True))
        batch_op.add_column(
            sa.Column('country', sa.String(3), nullable=True))
        batch_op.add_column(
            sa.Column('device', sa.String(80), nullable=True))
        batch_op.add_column(
            sa.Column('ip', sa.String(80), nullable=True))
        batch_op.add_column(
            sa.Column('os', sa.String(80), nullable=True))


def downgrade():
    """Downgrade database."""
    with op.batch_alter_table('accounts_user_session_activity') as batch_op:
        batch_op.drop_column('os')
        batch_op.drop_column('ip')
        batch_op.drop_column('device')
        batch_op.drop_column('country')
        batch_op.drop_column('browser_version')
        batch_op.drop_column('browser')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test password hashing."""

from binascii import hexlify, unhexlify

import flask
import pytest
from flask_security.utils import hash_password, verify_password

from invenio_accounts.hash import _to_binary, _to_string, mysql_aes_decrypt, \
    mysql_aes_encrypt
from invenio_accounts.testutils import client_authenticated, \
    create_test_user, login_user_via_view


def test_mysql_aes_encrypt():
    """Test mysql_aes_encrypt."""
    assert hexlify(mysql_aes_encrypt("test", "key")) == \
        b'9e9ce44cd9df2b201f51947e03bccbe2'
    assert hexlify(mysql_aes_encrypt(u"test", "key")) == \
        b'9e9ce44cd9df2b201f51947e03bccbe2'
    assert hexlify(mysql_aes_encrypt("test", u"key")) == \
        b'9e9ce44cd9df2b201f51947e03bccbe2'
    assert hexlify(mysql_aes_encrypt(u"test", u"key")) == \
        b'9e9ce44cd9df2b201f51947e03bccbe2'
    pytest.raises(AssertionError, mysql_aes_encrypt, object(), "key")
    pytest.raises(AssertionError, mysql_aes_encrypt, "val", object())


def test_mysql_aes_decrypt():
    """Test mysql_aes_encrypt."""
    assert mysql_aes_decrypt(
        unhexlify(b'9e9ce44cd9df2b201f51947e03bccbe2'), "key") == "test"
    assert mysql_aes_decrypt(
        unhexlify(u"9e9ce44cd9df2b201f51947e03bccbe2"), u"key") == "test"
    pytest.raises(AssertionError, mysql_aes_decrypt, object(), "key")
    pytest.raises(AssertionError, mysql_aes_decrypt, "val", object())


def test_context(app):
    """Test passlib password context."""
    with app.app_context():
        ctx = flask.current_app.extensions['security'].pwd_context
        hashval = hash_password("test")
        assert hashval != "test"
        assert verify_password("test", hashval)
        assert not ctx.needs_update(hashval)
        assert ctx.hash("test") != ctx.hash("test")


def test_conversion():
    """Test conversion between bytes and text."""
    str = 'abcd1234'
    byte_str = b'abcd1234'
    u_str = u'abcd1234'
    assert _to_binary(str) == byte_str
    assert _to_string(byte_str) == str
    assert _to_string(str) == str
    assert _to_binary(byte_str) == byte_str
    assert _to_string(u_str) == u_str


def test_unicode_regression(app):
    """Test legacy encryption with Unicode encoding."""
    with app.app_context():
        user = create_legacy_user(password=u'κωδικός')
        assert verify_password(u'κωδικός', user.password)


def test_invenio_aes_encrypted_email(app):
    """Test legacy encryption."""
    with app.app_context():
        ctx = flask.current_app.extensions['security'].pwd_context
        user = create_legacy_user()
        assert verify_password(user.password_plaintext, user.password)
        assert ctx.needs_update(user.password)


def test_user_login(app):
    """Test users' high-level login process."""
    with app.app_context():
        user = create_test_user('test@example.org')
        with app.test_client() as client:
            login_user_via_view(client, user.email, user.password_plaintext)
            assert client_authenticated(client)


def test_legacy_user_login(app):
    """Test legacy users' high-level login process."""
    with app.app_context():
        user = create_legacy_user()
        old_hashval = user.password
        with app.test_client() as client:
            login_user_via_view(client, user.email, user.password_plaintext)
            # Verify user is authenticated
            assert client_authenticated(client)
            # Verify password hash is upgraded
            ds = flask.current_app.extensions['security'].datastore
            user2 = ds.find_user(email=user.email)
            assert old_hashval != user2.password


def test_monkey_patch(app):
    """Test monkey-patching of Flask-Security's get_hmac function."""
    with app.app_context():
        user = create_test_user('test@example.org')
        assert verify_password(user.password_plaintext, user.password)


def test_monkey_patch_legacy(app):
    """Test monkey-patching of Flask-Security's get_hmac function."""
    with app.app_context():
        user = create_legacy_user()
        assert verify_password(user.password_plaintext, user.password)


def create_legacy_user(email='test@example.org', password=u'qwert1234',
                       **kwargs):
    """Create a legacy user in the datastore.

    A legacy user's password has been encrypted with the legacy mechanism,
    namely 'invenio_aes_encrypted_email'.
    """
    assert flask.current_app.testing
    ds = flask.current_app.extensions['security'].datastore

    # Encrypt password
    ctx = flask.current_app.extensions['security'].pwd_context
    encrypted_password = ctx.encrypt(
        password,
        scheme="invenio_aes_encrypted_email",
        salt=email,
    )

    # Update datastore
    user = ds.create_user(email=email, password=encrypted_password, **kwargs)
    ds.commit()
    user.password_plaintext = password

    return user



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test invenio-accounts models."""

from invenio_db import db
from sqlalchemy import inspect

from invenio_accounts import testutils
from invenio_accounts.models import SessionActivity


def test_session_activity_model(app):
    """Test SessionActivity model."""
    with app.app_context():
        # SessionActivity table is in the database
        inspector = inspect(db.engine)
        assert 'accounts_user_session_activity' in inspector.get_table_names()

        user = testutils.create_test_user('test@example.org')

        # Create a new SessionActivity object, put it in the db
        session_activity = SessionActivity(user_id=user.get_id(),
                                           sid_s="teststring")
        database = db

        # the `created` field is magicked in via the Timestamp mixin class
        assert not session_activity.created
        database.session.add(session_activity)
        # Commit it to the books.
        database.session.commit()
        assert session_activity.created
        assert len(user.active_sessions) == 1

        # Now how does this look on the user object?
        assert session_activity == user.active_sessions[0]

        session_two = SessionActivity(user_id=user.get_id(),
                                      sid_s="testring_2")
        database.session.add(session_two)
        # Commit it to the books.
        database.session.commit()

        assert len(user.active_sessions) == 2
        # Check #columns in table
        queried = database.session.query(SessionActivity)
        assert queried.count() == 2
        active_sessions = queried.all()
        assert session_activity.sid_s in [x.sid_s for x in active_sessions]
        assert session_two in queried.filter(
            SessionActivity.sid_s == session_two.sid_s)
        assert queried.count() == 2  # `.filter` doesn't change the query

        # Test session deletion
        session_to_delete = user.active_sessions[0]
        database.session.delete(session_to_delete)
        assert len(user.active_sessions) == 2  # Not yet updated.
        assert queried.count() == 1
        # Deletion is visible on `user` once database session is commited.
        database.session.commit()
        assert len(user.active_sessions) == 1
        assert user.active_sessions[0].sid_s != session_to_delete.sid_s
